/*
 * JBoss, Home of Professional Open Source
 * Copyright 2005, JBoss Inc., and individual contributors as indicated
 * by the @authors tag. See the copyright.txt in the distribution for a
 * full listing of individual contributors.
 *
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as
 * published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this software; if not, write to the Free
 * Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA, or see the FSF site: http://www.fsf.org.
 */
package org.jbpm.pvm.internal.identity.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;

import org.jboss.identity.idm.api.Attribute;
import org.jboss.identity.idm.api.AttributesManager;
import org.jboss.identity.idm.api.IdentitySearchCriteria;
import org.jboss.identity.idm.api.IdentitySession;
import org.jboss.identity.idm.api.RoleType;
import org.jboss.identity.idm.common.exception.IdentityException;
import org.jboss.identity.idm.common.p3p.P3PConstants;
import org.jboss.identity.idm.impl.api.SimpleAttribute;
import org.jbpm.api.JbpmException;
import org.jbpm.api.identity.Group;
import org.jbpm.api.identity.User;

/**
 * @author Tom Baeyens
 * @author Jeff Yu
 */
public class JBossIdmIdentitySessionImpl implements org.jbpm.pvm.internal.identity.spi.IdentitySession {

  protected IdentitySession identitySession;

  public JBossIdmIdentitySessionImpl(IdentitySession identitySession) {
    this.identitySession = identitySession;
  }

  public String createUser(String userName, String givenName, String familyName, String businessEmail) {
    try {
    	
      org.jboss.identity.idm.api.User idUser= identitySession.getPersistenceManager().createUser(userName);
      
      List<Attribute> attrs = new ArrayList<Attribute>();
      if (givenName != null) {
    	  attrs.add(new SimpleAttribute(P3PConstants.INFO_USER_NAME_GIVEN, givenName));
      }
      if (familyName != null) {
    	  attrs.add(new SimpleAttribute(P3PConstants.INFO_USER_NAME_FAMILY, familyName));
      }
      if (businessEmail != null) {
    	  attrs.add(new SimpleAttribute(P3PConstants.INFO_USER_BUSINESS_INFO_ONLINE_EMAIL, businessEmail));
      }
      
      
      identitySession.getAttributesManager().addAttributes(idUser, attrs.toArray(new Attribute[attrs.size()]));
      return idUser.getId();
      
    } catch (IdentityException e) {
      throw new JbpmException("couldn't create user "+userName, e);
    }

  }
  

  public List<User> findUsers() {
    try {
      Collection<org.jboss.identity.idm.api.User> idUsers = 
    	  identitySession.getPersistenceManager().findUser((IdentitySearchCriteria)null);    
      
      List<User> users = new ArrayList<User>();
      for (org.jboss.identity.idm.api.User idUser : idUsers) {
    	String name = idUser.getId();
        String givenName = getAttributeString(idUser, P3PConstants.INFO_USER_NAME_GIVEN);
        String familyName = getAttributeString(idUser, P3PConstants.INFO_USER_NAME_FAMILY);
        
        UserImpl user = new UserImpl(name, givenName, familyName);
        users.add(user);
      }
      
      return users;
      
    } catch (IdentityException e) {
      throw new JbpmException("couldn't get users from identity component", e);
    }
  }

  public User findUserById(String userId) {
	  try {
		org.jboss.identity.idm.api.User idUser = identitySession.getPersistenceManager().findUser(userId);
		if (idUser != null) {
	    	String name = idUser.getId();
	        String givenName = getAttributeString(idUser, P3PConstants.INFO_USER_NAME_GIVEN);
	        String familyName = getAttributeString(idUser, P3PConstants.INFO_USER_NAME_FAMILY);			
			
			User user = new UserImpl(name, givenName, familyName);
			return user;
		}
		return null;
	} catch (IdentityException e) {
		throw new JbpmException("couldn't get user from id of " + userId, e);
	}
  }

  public List<User> findUsersById(String... userIds) {
    List<User> users = new ArrayList<User>();
	for (String userId : userIds){
		User user = findUserById(userId);
		if (user != null) {
			users.add(user);
		}
	}
	  
    return users;
  }

  public List<User> findUsersByGroup(String groupId) {
	try {
		org.jboss.identity.idm.api.Group idGroup = findIdmGroupById(groupId);
		Collection<org.jboss.identity.idm.api.User> idusers = 
				identitySession.getRelationshipManager().findAssociatedUsers(idGroup, false);
		List<User> users = new ArrayList<User>();
		
		for (org.jboss.identity.idm.api.User iduser : idusers) {
			users.add(findUserById(iduser.getId()));
		}
		
		return users;
	} catch (IdentityException e) {
		throw new JbpmException("couldn't find users by groupid: " + groupId, e);
	}
	  
  }

  public void deleteUser(String userName) {
    try {    	
      identitySession.getPersistenceManager().removeUser(userName, true);    
    } catch (IdentityException e) {
      throw new JbpmException("couldn't delete user ["+userName + "]", e);
    }
  }

  public String createGroup(String groupName, String groupType, String parentGroupId) {
    
    try {
            	
      org.jboss.identity.idm.api.Group group = identitySession.getPersistenceManager().createGroup(groupName, groupType);
      String groupId = group.getId();
      
      if (parentGroupId!=null) {
        org.jboss.identity.idm.api.Group parentGroup = findIdmGroupById(parentGroupId);
        if (parentGroup==null) {
          throw new JbpmException("parent group "+parentGroupId+" doesn't exist");
        }
        identitySession.getRelationshipManager().associateGroups(parentGroup, group);
      }
     return groupId;
     
    } catch (IdentityException e) {
      throw new JbpmException("couldn't create group "+groupName, e);
    }
    
  }
  
  public Group findGroupById(String groupId) {
	try {
		org.jboss.identity.idm.api.Group idGroup = findIdmGroupById(groupId);
		GroupImpl group = new GroupImpl();
		group.setId(idGroup.getId());
		group.setType(idGroup.getGroupType());
		group.setName(idGroup.getName());
		
		Collection<org.jboss.identity.idm.api.Group> idParentGroups = 
			identitySession.getRelationshipManager().findAssociatedGroups(idGroup, null, false, false);
		
		if (idParentGroups.size() > 0) {
			org.jboss.identity.idm.api.Group idParent = idParentGroups.iterator().next();
			GroupImpl parentGroup = new GroupImpl();
			parentGroup.setId(idParent.getId());
			parentGroup.setType(idParent.getGroupType());
			parentGroup.setName(idParent.getName());
			
			group.setParent(parentGroup);
		}
		
		return group;
		
	} catch (IdentityException e) {
		throw new JbpmException("couldn't find group by id [" + groupId + "]", e);
	}  
  }

  public List<Group> findGroupsByGroupType(String groupType) {
    try {
     Collection<org.jboss.identity.idm.api.Group> idGroups = identitySession.getPersistenceManager().
     															findGroup(groupType);	
    
      List<Group> groups = new ArrayList<Group>();
      for (org.jboss.identity.idm.api.Group idGroup: idGroups) {
        groups.add(findGroupById(idGroup.getId()));
      }
      
      return groups;
      
    } catch (IdentityException e) {
      throw new JbpmException("couldn't get groups from identity component, groupType [" + groupType + "]", e);
    }
  }
  
  public List<Group> findGroupsByUser(String userId) {
	  
    try {
		Collection<org.jboss.identity.idm.api.Group> idGroups = identitySession.getRelationshipManager().
																	findAssociatedGroups(userId, null);
		
		List<Group> groups = new ArrayList<Group>();
		for (org.jboss.identity.idm.api.Group idGroup : idGroups) {
			groups.add(findGroupById(idGroup.getId()));
		}
		return groups;
	} catch (IdentityException e) {
		throw new JbpmException("Couldn't get Groups by userId [" + userId + "]", e);
		
	}
  }

  public List<Group> findGroupsByUserAndGroupType(String userName, String groupType) {
    try {
      org.jboss.identity.idm.api.User idUser = identitySession.getPersistenceManager().findUser(userName);
      
      Collection<org.jboss.identity.idm.api.Group> idGroups = identitySession.getRoleManager()
      														.findGroupsWithRelatedRole(idUser, groupType, null);
      List<Group> groups = new ArrayList<Group>();
      
      for (org.jboss.identity.idm.api.Group idGroup : idGroups) {
    	  groups.add(findGroupById(idGroup.getId()));
      }
      return groups;
    } catch (Exception e) {
      throw new JbpmException("couldn't get groups for user "+userName+" and groupType "+groupType, e);
    }
  }

  public void deleteGroup(String groupId) {
    try {
      org.jboss.identity.idm.api.Group group = findIdmGroupById(groupId);

      if (group==null) {
        return;
      }

      identitySession.getPersistenceManager().removeGroup(group, true);
    
    } catch (IdentityException e) {
      throw new JbpmException("couldn't delete group "+groupId, e);
    }
  }

  public void createMembership(String userId, String groupId, String role) {
    try {
      org.jboss.identity.idm.api.Group group = findIdmGroupById(groupId);
      if (group==null) {
        throw new JbpmException("group "+groupId+" doesn't exist");
      }
  
      org.jboss.identity.idm.api.User idUser = identitySession.getPersistenceManager().findUser(userId);
      if (idUser==null) {
        throw new JbpmException("user "+userId+" doesn't exist");
      }
      
      
      if (role!=null) {
        RoleType roleType = identitySession.getRoleManager().getRoleType(role);
        if (roleType==null) {
          roleType = identitySession.getRoleManager().createRoleType(role);
        }
        
        identitySession.getRoleManager().createRole(roleType, idUser, group);
        
      } else {
        identitySession.getRelationshipManager().associateUser(group, idUser);
      }
      
    } catch (Exception e) {
      throw new JbpmException("couldn't create membership "+userId+", "+groupId+", "+role, e);
    }
  }
  
  public void deleteMembership(String userId, String groupId, String role) {
	try {
		RoleType rtype = identitySession.getRoleManager().getRoleType(role);
		identitySession.getRoleManager().removeRole(rtype.getName(), userId, groupId);
	} catch (Exception e) {
		throw new JbpmException("couldn't delete the membership [" + userId + "," + groupId + "," + role + "]", e);
	}  
	
  }

  protected org.jboss.identity.idm.api.Group findIdmGroupById(String groupId) {
	try {
		Collection<org.jboss.identity.idm.api.Group> groups = identitySession.getPersistenceManager().findGroup(Group.TYPE_UNIT);
		for (org.jboss.identity.idm.api.Group group : groups) {
			if (groupId.equals(group.getId())) {
				return group;
			}
		}
		return null;
	} catch (IdentityException e) {
		throw new JbpmException("couldn't find the group by groupId: " + groupId, e);
	}  
  }

  protected String getAttributeString(org.jboss.identity.idm.api.User idUser, String attributeName) throws IdentityException {
    return getAttributeString(idUser, null, attributeName);
  }

  protected String getAttributeString(org.jboss.identity.idm.api.Group idGroup, String attributeName) throws IdentityException {
    return getAttributeString(null, idGroup, attributeName);
  }

  protected String getAttributeString(org.jboss.identity.idm.api.User idUser, org.jboss.identity.idm.api.Group idGroup, String attributeName) throws IdentityException {
    AttributesManager attributesManager = identitySession.getAttributesManager();
    Attribute attribute = null;
    if (idUser !=null) {
      attribute = attributesManager.getAttribute(idUser, attributeName);
    } else {
      attribute = attributesManager.getAttribute(idGroup, attributeName);
    }
    if (attribute!=null) {
      return (String) attribute.getValue();
    }
    return null;
  }


  public IdentitySession getIdentitySession() {
    return identitySession;
  }
  
  public void setIdentitySession(IdentitySession identitySession) {
    this.identitySession = identitySession;
  }


}
